#include <bt/bt_defs.h>
#include <bt/bt_hci_defs.h>
#include <bt/bt_hci_cmds.h>
#include <bt/bt_hci_dispatch.h>
#include <bt/bt_hci_api.h>
#include <sys/mutex.h>
#include <sys/heap.h>
#include <sys/timer.h>
#include <debug/logging.h>
#include <debug/toolbox.h>
#include <string.h>

// debug settings
#define LOG_LEVEL 0
#define LOG_CLASS 2
#include <debug/log_set.h>

#if LOG_LEVEL >= 4
char acl_dbg_buf[121*3];
#endif

/**
 * Returns the con handle a packet was received on.
 */
inline bt_hci_con_handle_t bt_acl_get_con_handle(struct bt_hci_pkt_acl* pkt)
{
	return (((u_char*)pkt)[2] & 0xF) << 8 | ((u_char*)pkt)[1];
}

/**
 * Returns the pb flag of a received packet
 */
inline u_char bt_acl_get_pb_flag(struct bt_hci_pkt_acl* pkt)
{
	return (((u_char*)pkt)[1] >> 4) & 3;
}

/**
 * Returns the bc flag of a received acl packet
 */
inline u_char bt_acl_get_bc_flag(struct bt_hci_pkt_acl* pkt)
{
	return (((u_char*)pkt)[1] >> 6) & 3;
}

/**
 * Returns a pointer to the payload of an ACL packet
 */
inline u_char* bt_acl_get_data_pointer(struct bt_hci_pkt_acl* pkt)
{
	return (pkt->payload);
}

inline u_short bt_acl_get_pkt_types(struct btstack* stack)
{
	return stack->acl_pkt_types;
}

inline u_short bt_acl_get_max_pkt_len(struct btstack* stack)
{
	return stack->host_acl_len;
}

// creates an ACL packet buffer
bt_acl_pkt_buf* bt_acl_pkt_buf_create(struct btstack* stack) {
	
	bt_acl_pkt_buf* buf;
	
	buf = NutHeapAllocClear(sizeof(bt_acl_pkt_buf));
	buf->pkt = NutHeapAllocClear(stack->host_acl_len + BT_ACL_HCI_PKT_HDR_LEN);
	return buf;
}

// destroys a single ACL packet buffer
void bt_acl_pkt_buf_destroy(bt_acl_pkt_buf* buf)
{
	NutHeapFree(buf->pkt);
	NutHeapFree(buf);
}

bt_acl_pkt_buf* bt_acl_pkt_queue_create(struct btstack* stack, u_short nr_bufs)
{
	u_short i;
	bt_acl_pkt_buf* que = NULL;
	bt_acl_pkt_buf* buf = NULL;
	// dyn. allocate packet buffers & add them to the queue of free buffers
	for (i=0; i<nr_bufs; i++) {
		// create a new msg fragment
		buf = bt_acl_pkt_buf_create(stack);
		buf->next = que;
		que = buf;
	}
	return que;
}

void bt_acl_pkt_queue_destroy(bt_acl_pkt_buf* que)
{
	bt_acl_pkt_buf* buf;
	
	while ((buf = que)) {
		// remove msg from queue
		que = buf->next;
		// destroy the msg
		bt_acl_pkt_buf_destroy(buf);
	}
}

/**
 * Opens an acl communication link.
 * 
 * returns 0 or error. a handle to the packet is stored in \param acl_pkt.
 */
long bt_acl_link_open(struct btstack* stack,
						bt_hci_con_handle_t hdl,
						u_char pb_flag,
						u_char bc_flag,
						u_short len,
						struct bt_hci_acl_link** ret_link)
{
	u_char hci_acl_hdr[HCI_ACL_DATA_PKT_HDR + 1];	// +1 for hci pkt type
	_bt_hci_module_con_handle_t module_hdl = BT_HCI_CON_HANDLE_OWN;
	long result = 0;
	u_char con = 0;
	
	struct bt_hci_acl_link* acl_link;
	*ret_link = NULL;
	
    
    if (hdl != BT_HCI_CON_HANDLE_OWN) {
	    // Get the external con_handle & increase the number of uncompleted packets
	    for (con = 0; con < BT_HCI_MAX_NUM_CON; con++) {
	        if (stack->connection[con].app_con_handle == hdl) {
	            break;
	        }
	    }
	    // The connection is invalid
	    if (con == BT_HCI_MAX_NUM_CON)
	        return BT_ERR_NO_CON;
	
	    // The connection exists in the table, but it is not in an open state!
	    if (stack->connection[con].state != BT_HCI_CONN_STATE_OPEN)
	        return BT_ERR_NO_CON;
	        
	    // else, store module con handle
	    module_hdl = stack->connection[con].module_con_handle;
    }
    
    // check payload size
    if (len > stack->host_acl_len)
    	return BT_ERR_ACL_PLOAD_OVFLW;

	// build hci packet header
	// Note: if self delivery (hdl == BT_HCI_CON_HANDLE_OWN) we don't do the
	// module_hdl <-> app_con_hdl conversion: in that case, module_hdl
	// is set to BT_HCI_CON_HANDLE_OWN (see bt_acl_link_close)
	hci_acl_hdr[0] = HCI_ACL_DATA_PACKET;
	hci_acl_hdr[1] = BTN_LO(module_hdl);
	hci_acl_hdr[2] = BTN_HI(module_hdl) | (pb_flag & 3) << 4 | (bc_flag & 3) << 6;
	hci_acl_hdr[3] = BTN_LO(len);
	hci_acl_hdr[4] = BTN_HI(len);

	// get access to the link
	acl_link = &stack->acl_link;
	_bt_semaphore_wait(&acl_link->sema);

	// if the packet has not to be sent to this device
	if (hdl != BT_HCI_CON_HANDLE_OWN) {	
		// wait until the bt module has completed enough packets
		_bt_semaphore_wait(&(stack->nr_acl_pkts));
		// increment number of completed packets for the given connection
		stack->connection[con].pkts_h2c += 1;
	    // acquire lock on uart
	    _bt_semaphore_wait(&(stack->transport.sem_write));	    
	    // send header
	    result = _UARTwrite(stack->transport.uart, hci_acl_hdr, sizeof(hci_acl_hdr));
	    // if error, release lock on uart and return
	    if (result < 0) {
		    _bt_semaphore_post(&(stack->transport.sem_write));
		    _bt_semaphore_post(&acl_link->sema);
		    return BT_ERR_UART_COM;
	    }
	}
	// self delivery!
	else {
		// Just copy the data to the acl pkt (see bt_stack)
		NutMutexLock(&stack->acl_pkt_mutex);
		memcpy(stack->acl_pkt, hci_acl_hdr, sizeof(hci_acl_hdr));
	}
    // else, give access to the link, and return
    acl_link->con_hdl = hdl;
    acl_link->written = 0;
    acl_link->data_len = len;
    (*ret_link) = acl_link;
    return 0;
}

/**
 * Closes a previously opened acl link.
 * 
 * Call this function after you've written your data to an acl link.
 * 
 * Note: does not close the baseband connection!
 */
void bt_acl_link_close(struct bt_hci_acl_link* acl_link)
{
	u_long t_stamp;
    
    // get pointer to the bt stack
    struct btstack* stack = acl_link->stack;
    
	// if this was not a self delivery, release lock on bt uart
	if (acl_link->con_hdl != BT_HCI_CON_HANDLE_OWN) {
		// release lock on uart
    	_bt_semaphore_post(&(stack->transport.sem_write));
    	// release acl link
    	_bt_semaphore_post(&acl_link->sema);
	}
    // else, we have to deliver the packet to this device
    else {
    	// release acl link
    	_bt_semaphore_post(&acl_link->sema);

    	DEBUG_STR("self delivery (%d bytes)\n%s\n", acl_link->written + sizeof(struct bt_hci_pkt_acl) - 1,
    				dbg_str_to_hex_str(acl_dbg_buf, (u_char*)stack->acl_pkt, acl_link->written + sizeof(struct bt_hci_pkt_acl) - 1));
		// get local time
    	t_stamp = NutGetMillis();
    	// call callback
    	stack->acl_pkt = stack->hci_acl_cb(stack->hci_acl_cb_arg,
    										stack->acl_pkt,
    										bt_acl_get_con_handle(stack->acl_pkt),
    										bt_acl_get_pb_flag(stack->acl_pkt),
    										bt_acl_get_bc_flag(stack->acl_pkt),
    										acl_link->data_len,
    										t_stamp);
    	// release lock on acl pkt
    	NutMutexUnlock(&stack->acl_pkt_mutex);
    	// yield CPU for that the receiving service can start handling the pkt
    	NutThreadYield();
    }
}

/**
 * Writes data to a previously opened acl link.
 * 
 * Call #bt_acl_link_open() to open an acl communication link.
 * Write your data to the link using #bt_acl_link_write().
 * After sending your data, release the link using #bt_acl_link_close().
 */
long bt_acl_link_write(struct bt_hci_acl_link* acl_link, u_char* data, u_short len)
{
	struct btstack* stack;
	
	// get pointer to the bt-stack
	stack = acl_link->stack;
	
	// prevent payload overflow
	if (acl_link->written + len > acl_link->data_len) {
		bt_acl_link_close(acl_link);
		return BT_ERR_ACL_PLOAD_OVFLW;
	}
	
	// if not self-delivery
	if (acl_link->con_hdl != BT_HCI_CON_HANDLE_OWN) {
		// if error, return
		if (_UARTwrite(stack->transport.uart, data, len)) {
			bt_acl_link_close(acl_link);
			return BT_ERR_UART_COM;
		}
	}
	// if self delivery, just copy the data
	else {
		memcpy(stack->acl_pkt->payload + acl_link->written, data, len);
	}
	// remember nr of bytes written
	acl_link->written += len;
	return 0;
}

/**
 * Application interface for sending an acl packet
 */
long bt_acl_send_packet(struct btstack* stack,
						bt_hci_con_handle_t hdl,
						u_char pb_flag,
						u_char bc_flag,
						u_char* data,
						u_short len)
{
	long result;
	struct bt_hci_acl_link* link;
	
	result =
		bt_acl_link_open(stack, hdl, pb_flag, bc_flag, len, &link);
	if (result) return result;
	
    // send data
    result = bt_acl_link_write(link, data, len);
    if (result) return result;
    
    // complete the packet
    bt_acl_link_close(link);
    return result;
}

u_short bt_acl_init(struct btstack* stack, u_short pkt_types)
{
	u_short acl_size;
	
	// set max acl size
    acl_size = BT_ACL_PAYLOAD_SIZE_DM1;
    if (pkt_types & BT_HCI_PACKET_TYPE_DH1) acl_size = BT_ACL_PAYLOAD_SIZE_DH1;
    if (pkt_types & BT_HCI_PACKET_TYPE_DM3) acl_size = BT_ACL_PAYLOAD_SIZE_DM3;
    if (pkt_types & BT_HCI_PACKET_TYPE_DH3) acl_size = BT_ACL_PAYLOAD_SIZE_DH3;
    if (pkt_types & BT_HCI_PACKET_TYPE_DM5) acl_size = BT_ACL_PAYLOAD_SIZE_DM5;
    if (pkt_types & BT_HCI_PACKET_TYPE_DH5) acl_size = BT_ACL_PAYLOAD_SIZE_DH5;
    
    // store pkt types & max acl size (host)
    stack->acl_pkt_types = pkt_types;
    stack->host_acl_len = acl_size;
    
    DEBUG("ACL: set pkt types to %d!\n", stack->acl_pkt_types);
    return acl_size;
}
