/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 */
 
/**
 * @file bt_hci_dispatch.h
 *
 * \author Mathias Payer <payerm@student.ethz.ch>
 * 
 * \date 2004/04/17 18:15:09 
 *
 * @brief Defines some functions to "alloc" pkts and dispatch them to the
 * hci-controller
 * @see bt_hci_dispatch.c for implementation details and descriptions
 */
 
#include <bt/bt_hci_defs.h>
#include <bt/bt_defs.h>

/*******************************************************************************
 * _bt_hci_pkts_init
 ******************************************************************************/
/**
 * 
 * Initializes all the packets of a stack.
 *
 * @param stack The bt-stack to use.
 */
void _bt_hci_pkts_init(struct btstack *stack);


/*******************************************************************************
 * _bt_hci_setWaitQueue
 ******************************************************************************/
/**
 *
 * Saves a bt_hci_cmd_response in the bt-internal layout.
 * The calle can wait with his bt_hci_cmd_response for the event to come back!
 *
 * @param stack The bt-stack to use.
 * @param resp the pointer to bt_hci_cmd_response, that will be entered in the btstack-queue
 * @return 0 if there was a free slot, -1 on error
 */
int _bt_hci_setWaitQueue(struct btstack *stack, struct bt_hci_cmd_response *resp);

/*******************************************************************************
* _bt_hci_dumpWaitQueue
******************************************************************************/
/** 
 * dumps the bt hci cmd waitqueue to stdout
 * @param stack The bt-stack to use.
 */
void _bt_hci_dumpWaitQueue(struct btstack *stack);

/*******************************************************************************
 * _bt_hci_wakeup
 ******************************************************************************/
/**
 *
 * Wake all threads that wait on ogcocf & con_handle combination and
 * stores the response in all waiting bt_hci_cmd_response structs
 * The call can wait with his bt_hci_cmd_response for the event to come back!
 *
 * @param stack The bt-stack to use.
 * @param ogcocf Which ocf/ogf event will be woken up
 * @param cmd_handle the cmd handle (BT_HCI_HANDLE_INVALID if unused)
 * @param response The response that the woken-up-thread gets back
 * @return 0 if an entry was fuond, -1 on error
 */
int _bt_hci_wakeup(struct btstack *stack, u_short ogcocf, u_short cmd_handle, long response);

/*******************************************************************************
* _bt_hci_wakeup_index
******************************************************************************/
/**
*
 * Wakes one thread of waitque and
 * stores the response in all waiting bt_hci_cmd_response structs
 * The call can wait with his bt_hci_cmd_response for the event to come back!
 *
 * @param stack The bt-stack to use.
 * @param index The index of bt_hci_cmd_response  to use.
 * @param response The response that the woken-up-thread gets back
 * @return 0 if an entry was fuond, -1 on error
 */
void _bt_hci_wakeup_index(struct btstack *stack, u_char index, long response);

/*******************************************************************************
 * _bt_hci_wakeup_latest
 ******************************************************************************/
/**
 *
 * Wakes thread that is waiting for Command Status Event and
 * stores the response in all waiting bt_hci_cmd_response structs
 * The call can wait with his bt_hci_cmd_response for the event to come back!
 *
 * @param stack The bt-stack to use.
 * @param index The index of bt_hci_cmd_response  to use.
 * @return 0 if an entry was fuond, -1 on error
 */
void _bt_hci_wakeup_latest(struct btstack *stack, long response);

/*******************************************************************************
 * _bt_hci_wakeup_unset
 ******************************************************************************/
/**
 * Sets latest wait queue command entry to NONE
 */
inline void _bt_hci_wakeup_unset(struct btstack *stack);

/*******************************************************************************
 * _bt_hci_cmd_get_response
 ******************************************************************************/
/**
 *
 * Saves a bt_hci_cmd_response in the bt-internal layout.
 * The calle can wait with his bt_hci_cmd_response for the event to come back!
 *
 * @param stack The bt-stack to use.
 * @param ogcocf Which ocf/ogf event will be woken up
 * @param cmd_handle The cmd handle (BT_HCI_HANDLE_INVALID if unused)
 * @return 0 on error, or the bt_hci_cmd_response struct otherwise
 */
struct bt_hci_cmd_response *_bt_hci_cmd_get_response(struct btstack *stack, u_short ogcocf, u_short cmd_handle);


/*******************************************************************************
 * _bt_hci_get_app_con_handle (module_con_handle -> app_con_handle)
 ******************************************************************************/
/**
 * Translates a module con handle into the app (unique) con handle
 *
 * @param stack The bt-stack to use.
 * @param module_con_handle The module con_handle
 *
 * @return BT_HCI_HANDLE_INVALID on error, the app_con_handle on success
 */
bt_hci_con_handle_t _bt_hci_get_app_con_handle(struct btstack *stack, _bt_hci_module_con_handle_t module_con_handle);


/*******************************************************************************
 * _bt_hci_get_module_con_handle (app_con_handle -> module_con_handle)
 ******************************************************************************/
/**
 * Translates a app (unique) con handle into the module con handle
 *
 * @param stack The bt-stack to use.
 * @param app_con_handle The module con_handle
 *
 * @return BT_HCI_HANDLE_INVALID on error, the module_con_handle on success
 */
_bt_hci_module_con_handle_t _bt_hci_get_module_con_handle(struct btstack *stack, bt_hci_con_handle_t app_con_handle);


/*******************************************************************************
 * _bt_hci_get_free_con_entry_index
 ******************************************************************************/
/**
 *
 * Gets an index of a free connection entry from the stack
 *
 * @param stack The bt-stack to use.
 * @return -1 on error, or the index otherwise
 */
short _bt_hci_get_free_con_entry_index(struct btstack *stack);

/*******************************************************************************
 * _bt_hci_intro_cmd
 ******************************************************************************/
/**
 *
 * Gets called by all hci cmds.
 * Waits until the controller accepts more pkts and the cmd-buffer is empty
 * Then it checks, if the call is sync or async.
 *
 * @param stack The bt-stack to use.
 * @param wcmd The local cmd_response, if the call is sync
 * @param response Pointer to the response.
 * @param ogf OGF
 * @param ocf OCF
 *
 * @return Pointer to the actual (local or remote) bt_hci_cmd_response that is used
 */
struct bt_hci_cmd_response *_bt_hci_intro_cmd(struct btstack *stack, struct bt_hci_cmd_response *wcmd, struct bt_hci_cmd_response *response, struct bt_hci_pkt_cmd *cmd, u_char ogf, u_short ocf);


/*******************************************************************************
 * _bt_hci_outro_cmd
 ******************************************************************************/
/**
 *
 * Gets called by all hci cmds.
 * If necessary, setup wcmdpointer struct with ocgfocf + connection handle,
 * then register the wcmdpointer in the stack's waitqueue.
 * Sends the pkt. to the transport
 * If call is synchronous (response == BT_HCI_SYNC), wait for cmd completion
 *
 * @param stack The bt-stack to use.
 * @param wcmdpointer Pointer to the used bt_hci_cmd_response
 * @param response Pointer to the response.
 * @param ogfocf OGF/OCF of this call
 * @param cmd_handle Handle of this call (app con handle used!!, BT_HCI_HANDLE_INVALID if unused)
 *
 * @return Pointer to the actual (local or remote) bt_hci_cmd_response that is used
 */
long _bt_hci_outro_cmd(struct btstack *stack, struct bt_hci_cmd_response *wcmdpointer, struct bt_hci_cmd_response *response, u_short ogfocf, u_short cmd_handle);

