/*
 * Copyright (C) 2000-2004 by ETH Zurich
 * Copyright (C) 2004 by Ole Reinhardt <ole.reinhardt@embedded-it.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */


/**
 * @file bt_l2cap.h
 *\author Ole Reinhardt <ole.reinhardt@embedded-it.de>
 *\date 25.01.2005
 * 
 * @brief Implements L2CAP communicarion layer
 * 
 * See Bluetooth 1.1 specification for better understanding of l2cap.
 */

#ifndef _BT_L2CAP_H_
#define _BT_L2CAP_H_

#include <sys/types.h>
#include <bt/bt_defs.h>

/*************************************************************/
/*! \name Some defines for error or not initialized conditions
 *
 */
/*************************************************************/

/** 
 *  Defines maximum nr of services the l2cap layer can handle.
 *  A service is the logical endpoint that will receive and handle
 *  the data fom the l2cap connection. In other words the next higher 
 *  layer (i.e. rfcomm)
 *
 *  Not more than 5 bits as they are used to encode the service nr in the unique l2cap id
 */
 
#define BT_L2CAP_MAX_CHANNELS      0xFE //253 
#define BT_L2CAP_MAX_SERVICES      16   // !!! not more than 5 bits as they are used to encode the service nr in the unique l2cap id
#define BT_L2CAP_HANDLE_INVALID    0x00
#define BT_L2CAP_CHNR_INVALID      0xFF
#define BT_L2CAP_SERVICENR_INVALID 0xFF
#define BT_L2CAP_CHANNEL_INFO_SIZE (sizeof(struct bt_l2cap_channel_info_data))

/*********************************/
/*! \name Header length defintions
 *
 */
/*********************************/


/**
 *  Header length of an acl packet
 */

#define BT_L2CAP_ACL_PKT_HDR_LEN   (5 + HCI_ACL_DATA_PKT_HDR) // 5 = BT_L2CAP_HEADER_LEN + 1 byte for the type. see bt_l2cap_acl_pkt

/** 
 *  Header length of a l2cap packet
 */

#define BT_L2CAP_HEADER_LEN         4

/** 
 *  Header length of a l2cap signaling packet
 */

#define BT_L2CAP_SIG_HEADER_LEN     4

/*************************/
/*! \name MTU definitions
 *
 */
/*************************/


/**
 *  Maximum transmission unit default values
 */

#define BT_L2CAP_MTU_DEFAULT        672
#define BT_L2CAP_MIN_MTU            48

/****************************************************/
/*! \name ACL packet sizes for differen acl pkt types
 *
 */
/****************************************************/

#define BT_L2CAP_ACL_SIZE_DM1       18
#define BT_L2CAP_ACL_SIZE_DH1       28
#define BT_L2CAP_ACL_SIZE_DM3       123
#define BT_L2CAP_ACL_SIZE_DH3       185
#define BT_L2CAP_ACL_SIZE_DM5       226
#define BT_L2CAP_ACL_SIZE_DH5       341

/**********************************/
/*! \name Signalling timeouts in ms
 *
 */
/**********************************/

#define BT_L2CAP_SIGNAL_TIMEOUT     5000 
#define BT_L2CAP_SIGNAL_TIMEOUT_PENDING 60000 // one minute if connection pending

/*********************************************/
/*! \name Predefined channel identifiers (CIDs)
 *
 */
/*********************************************/

#define BT_L2CAP_CID_USER           0x0040
#define BT_L2CAP_CID_SIG            0x0001
#define BT_L2CAP_CID_CONNECTIONLESS 0x0002
#define BT_L2CAP_CID_SERVICES       0xFF00
#define BT_L2CAP_CID_MAX            0xFFFF

/****************************************************/
/*!\name Signalling codes for l2cap signaling packets
 *
 */
/****************************************************/

#define BT_L2CAP_SIG_CMD_REJ  0x01
#define BT_L2CAP_SIG_CON_REQ  0x02
#define BT_L2CAP_SIG_CON_RSP  0x03
#define BT_L2CAP_SIG_CONF_REQ 0x04
#define BT_L2CAP_SIG_CONF_RSP 0x05
#define BT_L2CAP_SIG_DISC_REQ 0x06
#define BT_L2CAP_SIG_DISC_RSP 0x07
#define BT_L2CAP_SIG_ECHO_REQ 0x08
#define BT_L2CAP_SIG_ECHO_RSP 0x09
#define BT_L2CAP_SIG_INFO_REQ 0x0a
#define BT_L2CAP_SIG_INFO_RSP 0x0b

/***********************************************/
/*! \name Reject reasons of a signalling command
 *
 */
/***********************************************/

#define BT_L2CAP_REJ_CMD_NOT_UNDERSTOOD 0x0000
#define BT_L2CAP_REJ_MTU_EXCEEDED       0x0001
#define BT_L2CAP_REJ_INVALID_CID        0x0002

/************************************************/
/*! \name Types of information packets (SIG_INFO)
 *
 */
/************************************************/

#define BT_L2CAP_INFO_CONLESS_MTU       0x0001
#define BT_L2CAP_INFO_EXT_FEATURES      0x0002

#define BT_L2CAP_INFO_SUCCESS           0x0000
#define BT_L2CAP_INFO_NOT_SUPPORTED     0x0001

/*****************************************************/
/*! \name Channel state codes (internal state machine)
 *
 */
/*****************************************************/

#define BT_L2CAP_CH_NOT_ASSIGNED             0x00
#define BT_L2CAP_CH_CLOSED                   0x01
#define BT_L2CAP_CH_W4_L2CAP_CONNECT_RSP     0x02
#define BT_L2CAP_CH_W4_L2CAP_CONNECT         0x03
#define BT_L2CAP_CH_CONFIG                   0x04
#define BT_L2CAP_CH_OPEN                     0x06
#define BT_L2CAP_CH_W4_L2CAP_DISCONNECT_RSP  0x07

/***********************************/
/*! \name Some known predifined PSMs
 *
 */
/***********************************/

#define BT_L2CAP_PSM_INVALID             0x0000
#define BT_L2CAP_PSM_RFCOMM              0x0003
#define BT_L2CAP_PSM_BNEP                0x000F

/*********************************************/
/*! \name Connection request results (SIG_CON)
 *
 */
/*********************************************/

#define BT_L2CAP_CON_SUCCESS             0x0000
#define BT_L2CAP_CON_PENDING             0x0001
#define BT_L2CAP_CON_PSM_NOT_SUPPORTED   0x0002
#define BT_L2CAP_CON_SECURITY_BLOCK      0x0003
#define BT_L2CAP_CON_REJ_NO_RESSOURCE    0x0004

/*************************************************/
/*! \name Configuration request options (SIG_CONF)
 *
 */
/*************************************************/

#define BT_L2CAP_CONF_OPT_NONE           0x00
#define BT_L2CAP_CONF_OPT_MTU            0x01
#define BT_L2CAP_CONF_OPT_FLUSH          0x02
#define BT_L2CAP_CONF_OPT_QOS            0x03
#define BT_L2CAP_CONF_OPT_RTR_FLOW       0x04

/*************************************************/
/*! \name Configuration request results (SIG_CONF)
 *
 */
/*************************************************/

#define BT_L2CAP_CONF_SUCCESS            0x0000
#define BT_L2CAP_CONF_UNACCAPTABLE       0x0001
#define BT_L2CAP_CONF_REJECT             0x0002
#define BT_L2CAP_CONF_UNKNOWN_OPTIONS    0x0003

/**********************************************/
/*! \name Quality of service options (SIG_CONF)
 *
 */
/**********************************************/

#define BT_L2CAP_QOS_NO_TRAFFIC          0x00
#define BT_L2CAP_QOS_BEST_EFFORT         0x01
#define BT_L2CAP_QOS_GUARANTEED          0x02

/*********************************************************/
/*! \name Flow control and retransmission modes (SIG_CONF)
 *
 */
/*********************************************************/

#define BT_L2CAP_MODE_BASIC              0x00
#define BT_L2CAP_MODE_RTR                0x01
#define BT_L2CAP_MODE_FLOW               0x02

/*********************************/
/*! \name Connection status infos
 *
 */
/*********************************/

#define BT_L2CAP_NO_FURTHER_INFO         0x0000
#define BT_L2CAP_AUTHENTICATION_PENDING  0x0001
#define BT_L2CAP_AUTHORIZATION_PENDING   0x0002

/**********************************/
/*! \name Flags for channel status
 *
 */
/**********************************/

#define BT_L2CAP_FLAG_DISC_RSP    0x01
#define BT_L2CAP_FLAG_CONF_REQ    0x02
#define BT_L2CAP_FLAG_CONF_RSP    0x04   // Config response initiator
#define BT_L2CAP_FLAG_CON_RSP     0x08

#define BT_L2CAP_FLAG_WAS_CON     0x20
#define BT_L2CAP_FLAG_INITIATOR   0x80

#define BT_L2CAP_FLAG_CMD_MASK    0x0F

/*********************/
/*! \name Error codes
 *
 */
/*********************/

#define BT_L2CAP_ERR_SUCCESS           0x00
#define BT_L2CAP_ERR_INVALID_PSM       0x01
#define BT_L2CAP_ERR_BB_CON_FAILED     0x02
#define BT_L2CAP_ERR_NO_RESSOURCE      0x03
#define BT_L2CAP_ERR_TIMEOUT           0x04
#define BT_L2CAP_ERR_INVALID_CID       0x05
#define BT_L2CAP_ERR_CH_NOT_OPEN       0x06
#define BT_L2CAP_ERR_CMD_PENDING       0x07
#define BT_L2CAP_ERR_IN_USE            0x08
#define BT_L2CAP_ERR_INVALID_HANDLE    0x09
#define BT_L2CAP_ERR_DATASIZE          0x0A
#define BT_L2CAP_ERR_OUT_OF_MEMORY     0x0B

#define BT_L2CAP_ERR_PING_PENDING      0x10
#define BT_L2CAP_ERR_PAYLOAD_LEN       0x11
#define BT_L2CAP_ERR_CON_PENDING       0x12
#define BT_L2CAP_ERR_PSM_NOT_SUPPORTED 0x13
#define BT_L2CAP_ERR_SECURITY_BLOCK    0x14
#define BT_L2CAP_ERR_CH_CONFIG         0x15

/************************/
/*! \name Callback types 
 *
 */
/************************/

#define BT_L2CAP_CONNECT           0x01
#define BT_L2CAP_DISCONNECT        0x02

/************************/
/*! \name Callback detail
 *
 */
/************************/

#define BT_L2CAP_UNKNOWN               0x01
#define BT_L2CAP_CON_ACTIVE            0x02
#define BT_L2CAP_CON_PASSIVE           0x03
#define BT_L2CAP_DISCON_LINK_LOST      0x04

/**
 * Receive callback definition
 *
 * This callback will be called whenever any user data is received by the 
 * l2cap layer on a specific PSM. This is specified by the service registration
 * routine.
 * 
 * @param pkt Pointer to the reveiced acl packet
 * @param service_nr Number of the service this callback is assigned to
 * @param channel_id Channel ID of the l2cap channel this callback is assigned to
 * @param arg Pointer to the callback argument. (see register service)
 */

#define BT_L2CAP_RCV_CB void (*rcv_cb) (struct bt_l2cap_acl_pkt *pkt, u_char service_nr, u_short channel_id, void *arg)
    
/**
 * Connection callback definition
 *
 * This callback will be called whenever a l2cap connection is established or 
 * closed on a specific PSM. This is specified by the service registration
 * routine.
 * 
 * @param type Type of callback: BT_L2CAP_CONNECT or BT_L2CAP_DISCONNECT
 * @param detail Information about the callbacks reason. Possible values:
 *        BT_L2CAP_UNKNOWN
 *        BT_L2CAP_CON_ACTIVE
 *        BT_L2CAP_CON_PASSIVE
 *        BT_L2CAP_DISCON_LINK_LOST
 * @param service_nr Number of the service this callback is assigned to
 * @param channel_id Channel ID of the l2cap channel this callback is assigned to
 * @param arg Pointer to the callback argument. (see register service)
 */

#define BT_L2CAP_CON_CB void (*con_cb) (u_char type, u_char detail, u_char service_nr, u_short channel_id, void *arg)

struct l2cap_echo_cmd {
    u_char id;
    HANDLE echo_ev;
};

/**
 * @struct bt_l2cap_acl_pkt
 * @brief Represents one l2cap packet including the acl header
*/

struct bt_l2cap_acl_pkt {
    u_char  type;			        /**< Type of the packet */
    u_char  header[HCI_ACL_DATA_PKT_HDR];	/**< ACL-Header, handled by send_acl_pkt */
    u_char  len[2];
    u_char  cid[2];
    u_char  payload[1];
} __attribute__ ((packed));

/**
 * @struct bt_l2cap_channel_info_data
 * @brief This struct contains some internal information to process the packets. This information
 *        is added at the end of a l2cap data packet.
 */

struct bt_l2cap_channel_info_data {
    u_short channel_id;                         // channel id to check if channel is still open
    u_char  buf_used;                           // l2cap buffer used or only direct acl buffer
    struct bt_l2cap_acl_pkt* pkt;               // pointer to last acl buffer
    struct bt_l2cap_channel* channel;           
} __attribute__ ((packed));

/**
 *  @struct bt_l2cap_pkt_buffer_entry
 *  @brief Buffer entry for internal l2cap packet buffering
 */

struct bt_l2cap_pkt_buffer_entry {
    struct  bt_l2cap_acl_pkt *pkt;
    u_char  state;
};

/** 
 *  @struct bt_l2cap_service
 *  @brief This struct represents one service. 
 * 
 *  It contains all the information needed to handle a service.
 *
 *  A service is the logical endpoint that will receive and handle
 *  the data fom the l2cap connection. In other words the next higher 
 *  layer (i.e. rfcomm). The l2cap layer can handle multiple services
 *  Each service can be connected to one channel.
 *  @see struct bt_l2cap_channel for more information about the channels
 *  @see struct bt_l2cap_stack for more information about the l2cap layer itself
 *
 */

struct bt_l2cap_service {
    u_char  channel_nr; // Number of the channel associated with this service
    bt_hci_con_handle_t con_handle; // connection handle of the baseband connection this service is associated to.
    u_short channel_id; // Unique ID passed to the upper layers. We identify our connection / channel, using this ID.
    u_short psm;        // This contains the local or remove psm depending on who initiated the connection
    u_short min_mtu;    // Minimum transmission unit we need to run our service
    u_short max_mtu;    // Maximum transmission unit we can handle with our service
    struct  bt_l2cap_pkt_buffer_entry *buffer; // Packet buffers belonging to the service
    u_char  buffer_size;  // size of the buffer above
    BT_L2CAP_RCV_CB;    // Receive callback
    BT_L2CAP_CON_CB;    // Connection callback
    void    *cb_arg;    // Callback argument.
};

/** 
 *  @struct bt_l2cap_channel
 *  @brief This struct represents one l2cap channel
 *  
 *  This struct contains all the information needed to handle a l2cap channel
 *  Every service can handle only one channel, but one baseband connection can
 *  open multiple channels to multiple sevices of the same device. 
 *  @see struct bt_l2cap_stack for more information about the stack
 *  @see struct bt_l2cap_service for more information about a service
 */
 
struct bt_l2cap_channel {
    u_char  state;   // State of channel... see state codes above
    bt_hci_con_handle_t con_handle;
    u_short remote_cid;
    u_short local_cid;
    u_char  flags;   // See flag definitions above
    u_char  sig_id;  // The id of the last send signal command
    u_short omtu;    // maximum output transmission unit
    u_short imtu;    // maximum incomming transmission unit (omtu of the other side)
    u_char  error;   // error code of the current operation
    u_char  pending; // 1 if connection pending... (PIN / Linkkey authentification)
    
    HANDLE  sig_con_rsp;
    HANDLE  sig_conf_rsp;
    HANDLE  sig_disc_rsp;
    HANDLE  sig_timeout;
    
    u_char  service_nr;
    u_short received;
    u_short buf_state;
    struct  bt_l2cap_acl_pkt *pkt;
    struct  bt_l2cap_acl_pkt *last_acl_pkt;
};

/** 
 *  @struct bt_l2cap_signal
 *  @brief This struct contains the header fields of a l2cap signaling packet
 */

struct bt_l2cap_signal {
    u_char  code;
    u_char  id;
    u_char  len[2];
    u_char  payload[1];
} __attribute__ ((packed));

/**
 *  @struct bt_l2cap_acl_buffer_entry
 *  @brief Buffer entry for internal acl packet buffering
 */

struct bt_l2cap_acl_buffer_entry {
    struct bt_l2cap_acl_pkt *pkt;
    bt_hci_con_handle_t con_handle;
    u_char  pb_flag;
    u_char  state;
    u_short len;
    u_short sequence_nr;
};

/**
 *  @struct bt_l2cap_signal_buffer
 *  @brief Buffer entry for internal l2cap signal buffering
 */

struct bt_l2cap_signal_buffer {
    struct bt_l2cap_acl_pkt *signal;
    u_short received;
    u_char  state;
};

/**
 *  @struct bt_l2cap_bb_connection
 *  @breif Represents one entity of a baseband connection.
 *
 *  The l2cap layer is able to handle multiple baseband connections.
 *  This struct defines all information needed to handle a baseband connection
 */

struct bt_l2cap_bb_connection {
    struct bt_l2cap_signal_buffer signal;
    struct bt_l2cap_channel **channels;
    u_char connections;
    u_char sig_id;
    u_long pkt_types;
    bt_hci_con_handle_t con_handle;
};

/**
 *  @struct bt_l2cap_stack
 *  @brief Represents one entity of the l2cap layer itself.
 *  
 *  This struct contains all the data needed by the l2cap layer itself.
 *  It holds multiple packet buffers, information about the connected
 *  channels and the registered services.
 */

struct bt_l2cap_stack {
    struct btstack *bt_stack;   // pointer to lower level hci stack             
    struct bt_l2cap_service services[BT_L2CAP_MAX_SERVICES];
    struct bt_l2cap_bb_connection bb_con[BT_HCI_MAX_NUM_CON];
    struct bt_l2cap_acl_buffer_entry *buffer;    //array of buffer_size
    struct bt_l2cap_acl_pkt *tmp_pkt;
    struct bt_l2cap_channel *channels;
    u_char buffer_size;
    u_char bb_connections;
    u_short pkt_types;
    u_short max_signal_mtu; // length of the payload field in bt_l2cap_pkt_t
    u_short acl_len;        // length of the payload field in bt_l2cap_acl_pkt_t   
    u_char  max_channel;    // maximum nr. of open l2cap channels
    u_short channel_id;
    u_short acl_sequence_nr;
    HANDLE  sig_acl_buffer_free;
    HANDLE  sig_acl_data_avail;
};

/* These functions are internal functions... */

u_char _bt_l2cap_get_new_sigid(struct bt_l2cap_bb_connection* bb_con);
struct bt_l2cap_bb_connection* _bt_l2cap_handle2baseband(bt_hci_con_handle_t con_handle);
struct bt_l2cap_service * _bt_l2cap_get_service_from_channel_id(u_short channel_id);
struct bt_l2cap_channel * _bt_l2cap_get_channel_from_channel_id(u_short channel_id);
    
/*******************************************************************************
 * bt_l2cap_register_service
 ******************************************************************************/
/**
 * @brief This command registers a service (the next higher layer)
 * 
 * A service is the logical endpoint that will receive and handle
 * the data fom the l2cap connection. In other words the next higher 
 * layer (i.e. rfcomm)
 * 
 * Changed MTU handling: Now we only have to define a minimum MTU the opposite device
 * has to accept for this service. Our maximum MTU is given as the stacks max_mtu
 *
 * @param psm The PSM the service will use
 * @param nr_buffer Number of l2cap packet buffers that should be reserved for this service
 *        Note: The service must be able to handle nr_buffer + number of acl 
 *        buffers given on l2cap initialisation l2cap packets before one packet is freed.
 *        So always keep in mind to reserve enough buffers in your service. See bt_rfcomm for example
 * @param min_mtu Minimum MTU value that the opposite device has to accept
 * @param max_mtu Maximum MTU value our service will use
 * @param BT_L2CAP_CON_CB Connection callback. Will be called if the datalink connection to 
 *        this service is established or closed.
 * @param BT_L2CAP_RCV_CB Receive callback. Will be called whenever data is received on an 
 *        established datalink
 * @param cb_arg Pointer to a optional callback argument
 * 
 * @return Number of the service or BT_L2CAP_SERVICENR_INVALID on an errors
 */

u_char bt_l2cap_register_service(u_short psm, u_char nr_buffer, u_short min_mtu, u_short max_mtu, BT_L2CAP_CON_CB, BT_L2CAP_RCV_CB, void* cb_arg);

/*******************************************************************************
 * bt_l2cap_clear_service
 ******************************************************************************/
/**
 * @brief This command clears a previosly registered service.
 * 
 * After calling this command, the service is no longer registered. The service
 * may not be in use.
 * 
 * @param service_nr Number of the service to be cleared
 *
 * @return BT_L2CAP_ERR_SUCCESS or one of the following errors:
 *   - BT_L2CAP_ERR_IN_USE
 */

u_char bt_l2cap_clear_service(u_char service_nr);

/*******************************************************************************
 * bt_l2cap_connect
 ******************************************************************************/
/**
 * @brief This command will open a channel to a remote device 
 * 
 * Opening a channel will connect a waiting service with the specified PSM
 * of the remote device. If no baseband connection is yet established, a baseband
 * connection will be established first. After successfull connection the data can be 
 * transferred. 
 *
 * Uppler layers will be informed by a connection callback
 *
 * @param service_nr Number of the service to be connected to the remote device
 * @param bt_addr Bluetooth address of the remote device
 * @param page_scan_rep_mode Page scan mode (found during inquiry) (pass 0 if unknown)
 * @param clock_offset Offset to this device (found during inquiry). pass 0 if unknown
 * @param psm PSM of the connection endpoint
 * @param channel_id of new connection is stored here, if not null
 *
 * @return BT_L2CAP_ERR_SUCCESS or one of the following errors:
 *   - BT_L2CAP_ERR_INVALID_HANDLE
 *   - BT_L2CAP_ERR_NO_RESSOURCE
 *   - BT_L2CAP_ERR_INVALID_PSM
 *   - BT_L2CAP_ERR_BB_CON_FAILED
 *   - BT_L2CAP_ERR_TIMEOUT
 */

u_char bt_l2cap_connect(u_char service_nr, bt_addr_t bt_addr, 
                        u_char page_scan_rep_mode, u_short clock_offset, 
                        u_short psm, u_short *channel_id);

/*******************************************************************************
 * bt_l2cap_disconnect
 ******************************************************************************/
/**
 * @brief This command will close a previously opened connection
 * 
 * After closing the channel no more data can be transfered. 
 * Uppler layers will be informed by a connection callback
 *
 * @param channel_id ID of the channel to be closed. Will be passed by the connection callback.
 *
 * @return BT_L2CAP_ERR_SUCCESS or one of the following errors:
 *   - BT_L2CAP_ERR_INVALID_HANDLE
 *   - BT_L2CAP_ERR_INVALID_CID
 *   - BT_L2CAP_ERR_CH_NOT_OPEN
 *   - BT_L2CAP_ERR_CMD_PENDING
 *   - BT_L2CAP_ERR_TIMEOUT
 */

u_char bt_l2cap_disconnect(u_short channel_id);

/*******************************************************************************
 * bt_l2cap_get_local_cid
 ******************************************************************************/
/**
 * @brief Queries the local cid of an established channel
 *
 * @param channel_id ID of the channel
 * @param cid Local CID of the channel (call by reference) 
 *
 * @return BT_L2CAP_ERR_SUCCESS or one of the following errors:
 *   - BT_L2CAP_ERR_INVALID_HANDLE
 */

u_char bt_l2cap_get_local_cid(u_short channel_id, u_short *cid);

/*******************************************************************************
 * bt_l2cap_get_remote_cid
 ******************************************************************************/
/**
 * @brief Queries the remote cid of an established channel
 *
 * @param channel_id ID of the channel
 * @param cid Remote CID of the channel (call by reference) 
 *
 * @return BT_L2CAP_ERR_SUCCESS or one of the following errors:
 *   - BT_L2CAP_ERR_INVALID_HANDLE
 */

u_char bt_l2cap_get_remote_cid(u_short channel_id, u_short *cid);

/*******************************************************************************
 * bt_l2cap_get_outgoing_mtu
 ******************************************************************************/
/**
 * @brief Queries the outgoing mtu of an established channel
 *
 * @param channel_id ID of the channel
 * @param mtu Outgoing MTU of the channel (call by reference) 
 *
 * @return BT_L2CAP_ERR_SUCCESS or one of the following errors:
 *   - BT_L2CAP_ERR_INVALID_HANDLE
 */

u_char bt_l2cap_get_outgoing_mtu(u_short channel_id, u_short *mtu);

/*******************************************************************************
 * bt_l2cap_get_incomming_mtu
 ******************************************************************************/
/**
 * @brief Queries the incomming mtu of an established channel
 *
 * @param channel_id ID of the channel
 * @param mtu Incomming MTU of the channel (call by reference) 
 *
 * @return BT_L2CAP_ERR_SUCCESS or one of the following errors:
 *   - BT_L2CAP_ERR_INVALID_HANDLE
 */

u_char bt_l2cap_get_incomming_mtu(u_short channel_id, u_short *mtu);

/*******************************************************************************
 * bt_l2cap_get_hci_connection_handle
 ******************************************************************************/
/**
 * @brief Queries the baseband connection handle an established channel
 *
 * @param channel_id ID of the channel
 * @param con_handle Baseband connection handle of the channel (call by reference) 
 *
 * @return BT_L2CAP_ERR_SUCCESS or one of the following errors:
 *   - BT_L2CAP_ERR_INVALID_HANDLE
 */

u_char bt_l2cap_get_hci_connection_handle(u_short channel_id, bt_hci_con_handle_t *con_handle);

/*******************************************************************************
 * bt_l2cap_is_ch_open
 ******************************************************************************/
/**
 * @brief Checks if a channel_id is valid (the channel is opened)
 *
 * @return TRUE if assigned, else false
 */ 
 
u_char bt_l2cap_is_ch_open(u_short channel_id); 

/*******************************************************************************
 * bt_l2cap_l2ping
 ******************************************************************************/
/**
 * @brief Send a l2cap echo request to a specific device and queries the 
 *        return time of the packet.
 *
 * @param bt_addr Bluetooth address of the remote device
 * @param payload Payload to be send with the echo request
 * @param len Length of the payload (my not be larger than one acl packet)
 * @param time returns the return time of the packet (call by reference)
 * @return BT_L2CAP_ERR_SUCCESS or one of the following errors:
 *   - BT_L2CAP_ERR_PAYLOAD_LEN
 *   - BT_L2CAP_ERR_PING_PENDING
 *   - BT_L2CAP_ERR_BB_CON_FAILED
 *   - BT_L2CAP_ERR_TIMEOUT
 */ 

u_char bt_l2cap_l2ping(bt_addr_t bt_addr, u_char *payload, u_short len, u_long *time);

/*******************************************************************************
 * bt_l2cap_send
 ******************************************************************************/
/**
 * @brief Sends data over a connected channel
 *
 * @param channel_id ID of the channel
 * @param pkt l2cap packet to be send. Fillin only the payload, header will 
 *        be added by the function.
 * @param len Length of the payload (may not be larger than the outgoing mtu of the channel
 * @return BT_L2CAP_ERR_SUCCESS or one of the following errors:
 *   - BT_L2CAP_ERR_INVALID_HANDLE
 *   - BT_L2CAP_ERR_DATASIZE
 *   - BT_L2CAP_ERR_INVALID_CID
 *   - BT_L2CAP_ERR_CH_NOT_OPEN
 *   - BT_L2CAP_ERR_CMD_PENDING
 *   - One of the errors from bt_hci_send_acl_pkt
 */

u_long bt_l2cap_send(u_short channel_id, struct bt_l2cap_acl_pkt* pkt, u_short len);

/*******************************************************************************
 * bt_l2cap_complete_pkt
 ******************************************************************************/
/**
 * @brief Informs the baseband that an packet can be freed (flow control)
 *
 * This function is used for flow control reasons. Every packet received by a service
 * needs to be completed with this function. This function will inform the baseband that
 * the reserved buffer can be cleared and will also clear l2cap internal buffers.
 * If the packets won't be completed the baseband will stop receiving any more packets after
 * The internal buffer runs full. 
 *
 * @param pkt l2cap packet to be completed
 */

void bt_l2cap_complete_pkt(struct bt_l2cap_acl_pkt *pkt);


/**
 * @bried Initializes the l2cap stack. 
 * 
 * @param bt_stack Pointer to btstack
 * @param nr_buffer Number of receive-buffer for l2cap packets
 * @param max_channel Maximum number of channels that can be used over one bb connection
 * @param pkt_types Bitmask of the acl packet types allowed. 
 *        Example:
 *        BT_HCI_PACKET_TYPE_DM1 | BT_HCI_PACKET_TYPE_DH1 | BT_HCI_PACKET_TYPE_DM3 | BT_HCI_PACKET_TYPE_DH3
 * @param mtu Maximum transmission unit for all channels (Needs be to equal or lager 
 *        than the max mtu needed by one of the services.
 *
 * @return Pointer to the l2cap stack entity created by this function.
 */

struct bt_l2cap_stack* bt_l2cap_init(struct btstack *bt_stack, u_char nr_buffer, u_char max_channel, u_short pkt_types);

extern struct l2cap_echo_cmd echo_cmd;
extern struct bt_l2cap_stack *l2cap_stack;

#endif
