/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file l2cap_cl.h
 * 
 * \author Kevin Martin <kevmarti@tik.ee.ethz.ch> 
 * 
 * \date 2006/03/23
 * 
 * \brief Connection-less data protocol API.
 * 
 * This is the API of the connection-less transport layer, which can be
 * used to send data to directely connected neighbors. It is the underlying
 * protocol used by the connection-less multi-hop protocol (see #mhop_cl.h)
 * to send data from hop to hop.
 * 
 * Actually, this is not an implementation of the l2cap connection-less
 * protocol as defined by the Bluetooth specification, although the packet
 * format used is equivalent to the definition of l2cap
 * connection-less packets (see note below).
 * 
 * - Sending packets is done by calling #l2cap_cl_send. E.g.
\verbatim
u_char data[12];
strcpy(data, "hello world");
l2cap_cl_send(data, <con_handle>, <psm>, 12);
\endverbatim
 * 
 * - each packet carries an identifier (the psm), which uniquely identifies
 * to which service the packet is delivered on the receiving side
 * 
 * - receiving packets, i.e. assigning a psm to your own service is done
 * by registering your service at the protocol/service multiplexor
 * (see #bt_psm_service_register).
 * 
 * Note: at that moment, it is not possible to run the connection-less data
 * protocol described here together with the current implementation of the
 * l2cap layer (see #l2cap.h).
 * 
 * \see #bt_psm.h #bt_l2cap.h #mhop_cl.h
 */
#ifndef _L2CAP_CL_H_
#define _L2CAP_CL_H_

#include <sys/types.h>
#include <bt/bt_defs.h>
#include <bt/bt_hci_defs.h>
#include <bt/bt_psm.h>

/// Length of the L2CAP connectionless packet header size
#define L2CAP_CL_PKT_HDR_LEN	(4 + L2CAP_CL_PSM_FIELD_SIZE)

/**
 * \brief Size of the PSM field.
 * 
 * See BT specification 1.2, Vol. 3, Part A, p. 34 for further details.
 */
#define L2CAP_CL_PSM_FIELD_SIZE					2

/**
 * The payload of the l2cap connection-less data packet to send is too large
 */
#define L2CAP_CL_ERR_PAYLOAD_OVERFLOW		   	-1

/**
 * Typedef of the handle used for accessing the l2cap connection-less channel.
 * 
 * \see #l2cap_cl_channel_open()
 */
typedef struct l2cap_cl_channel_s l2cap_cl_channel_t;

/**
 * \brief Sends a connection-less l2cap data packet.
 * 
 * Sends a connection-less l2cap data packet to the connection-less channel
 * (i.e. l2cap channel id equals 0x02) over the connection handle
 * \p con_handle to the psm \p psm.
 * 
 * Create your data, and send the packet to the connection with handle
 * 'con_handle' passing the pointer to your data, the psm the data shall be
 * sent to, as well as the length of the data to send.\n
 * e.g.:\n
\verbatim
u_char data[12];
strcpy(data, "hello world");
l2cap_cl_send(data, <con_handle>, <psm>, 12);
\endverbatim
 * Note: The maximal amount of data that can be obtained by calling
 * #l2cap_cl_max_payload.
 * 
 * \param data pointer to the data to be sent
 * \param len size of the data to be sent
 * \param hdl connection handle the data shall be sent to
 * \param psm identifies the service the data will be delivered to
 * on the receiving side (see #bt_psm_service_register).
 * \return 0 or possible errors: 
 * - #BT_ERR_BT_MODULE\n
 * - #BT_ERR_NO_CON\n
 * - #L2CAP_CL_ERR_ACL_PAYLOAD_OVERFLOW
 * 
 * \see bt_psm.h
 */
long l2cap_cl_send(u_char* data,
					u_short len,
					bt_hci_con_handle_t hdl,
					u_short psm);

/**
 * \brief Opens the l2cap connectionless channel.
 * 
 * Opens the l2cap connectionless channel for continiousely writing data to it.
 * 
 * Note: Provides unique access to the bluetooth device for writing data.
 * The lock on the bluetooth device is not released until
 * #l2cap_cl_channel_close has been called.
 * 
 * Applications typically do not call this function, but make use
 * of #l2cap_cl_send instead.
 * 
 * \param hdl Connection handle of the connected neighbor device the data
 * shall be sent to.
 * \param psm PSM the data will be sent to on the receiving side.
 * \param len Amount of data to be sent.
 * \param cl_chan_pp Pointer to a variable where the pointer to the opened
 * connection-less channel shall be written to. The pointer obtained then
 * serves as a handle for writing data to the connection-less channel.
 * 
 * \see #l2cap_cl_channel_write, #l2cap_cl_channel_close
 */
long l2cap_cl_channel_open(bt_hci_con_handle_t hdl,
							u_short psm,
							u_short len,
							l2cap_cl_channel_t** cl_chan_pp);

/**
 * \brief Closes the l2cap connectionless channel.
 * 
 * Closes the l2cap connection-less channel. Releases the writing lock on
 * the bluetooth controller triggered by #l2cap_cl_channel_open.
 * 
 * Note: Applications typically do not call this function, but make use
 * of #l2cap_cl_send instead.
 * 
 * \param chan Pointer to the previousely opened connection-less channel
 * 
 * \see #l2cap_cl_channel_open, #l2cap_cl_channel_write
 */
inline void l2cap_cl_channel_close(l2cap_cl_channel_t* chan);

/**
 * \brief Writes data to the l2cap connection-less channel
 * 
 * Writes data to the previousely opened connection-less data channel
 * (using #l2cap_cl_channel_open).
 * 
 * Note: Applications typically do not call this function, but make use
 * of #l2cap_cl_send instead.
 * 
 * \param chan Pointer to the connection-less channel
 * \param data Data to write
 * \param len Size of the data to write
 * 
 * \see #l2cap_cl_channel_open, #l2cap_cl_channel_close
 */
inline long l2cap_cl_channel_write(l2cap_cl_channel_t* chan, u_char* data, u_short len);

/**
 * \brief Returns a pointer to the payload of a received l2cap connection-less
 * packet.
 * 
 * \param pkt Pointer to the received packet
 */
inline u_char* l2cap_cl_get_data_ptr(struct bt_hci_pkt_acl* pkt);

/**
 * \brief Returns the max payload size of a connection-less l2cap packet.
 */
inline u_short l2cap_cl_max_payload(void);

inline u_long l2cap_cl_get_arrival_time(bt_msg* msg);

/**
 * \brief Initializes the connectionless l2cap stack. 
 * 
 * Note: The bt-stack, the ACL layer, as well as the protocol/service
 * multiplexor have to be initialized before calling this function. Thus, make
 * shure that:
 * - #bt_hci_init()
 * - #bt_acl_init()
 * - #bt_psm_init()
 * have been called.
 * 
 * \param bt_stack Pointer to the bt-stack to use
 * \param mux Pointer to the protocol/service multiplexor to use
 */
void l2cap_cl_init(struct btstack* bt_stack, bt_psm_t* mux);

/**
 * \brief ACL callback registration function
 * 
 * Call this function to register a higher layer service which handles the data
 * that is not handled by the l2cap connection-less layer (i.e. l2cap packets
 * that are not sent to the connection-less channel).
 * 
 * Note: l2cap_cl and l2cap cannot run together at the moment!
 */
void l2cap_cl_register_acl_cb(struct btstack* bt_stack, HCI_ACL_CB, void* arg);

#endif //_L2CAP_CL_H_
