/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

#ifndef TOOLBOX_H_
#define TOOLBOX_H_

#include <bt/bt_defs.h>

#define ADDR_FMT "%.4x:%.4x:%.4x"
#define ADDR(addr) \
    addr[5]<<8 | addr[4], addr[3]<<8 | addr[2], addr[1]<<8 | addr[0]
#define SADDR_FMT "%.2x:%.2x"
#define SADDR(addr) \
    addr[1], addr[0]

void bt_print_bt_addr(bt_addr_t addr);

char* dbg_str_to_hex_str(char* str, u_char* buf, u_short size);

/**
 * \brief Convert a bt_addr_t to a string.
 * 
 * Convert a bt_addr_t struct to a string of the form "xx:xx:xx:xx:xx:xx"
 * 
 * \param addr   BT address to be converted.
 * \return       Pointer to the stored string (same as buffer).
 */
char *dbg_bt_addr_to_str(bt_addr_t addr);

/**
 * \brief Convert Bluetooth address in string to bt_addr_t.
 * 
 * This function recognizes any optional spaces before the address.
 * The function returns a pointer to the first character in the string 
 * after the address.
 * If no address was found the function returns a pointer to the beginning 
 * of the string and addr will contain the contents of bt_addr_null. 
 * 
 * \param str The string to convert.
 * \param addr Address buffer.
 * \return Pointer to the first character not converted.
 */
char* string_to_bt_addr(char* str, char* addr);

/**
 * \brief Read user input from stream.
 *
 * Read at most one less than the specified number of characters from 
 * a stream or stop when a newline has been read. Echoes read chars to stream.
 * String in buffer is always null terminated.
 *
 * \param stream Pointer to a previously opened uart stream.
 * \param buffer Pointer to the buffer that receives the data
 * \param count  Maximum number of characters to read.
 * \return Length of input string in buffer.
 */
u_char user_input( FILE* stream, char *buffer, u_char count);

/**
 * \brief Get uart errors.
 * 
 * This function returns the current uart error flags (framing/overrun/parity) 
 * of the given stream. The returned flags are cleared. 
 * 
 * \param stream Pointer to a previously opened uart stream.
 *
 * \warning The function will not check, if the stream pointer points
 *          to a valid stream.
 */
u_char get_uart_errors(FILE* stream);

#endif /*TOOLBOX_H_*/
