#include <terminal/btn-terminal.h>
#include <eepromdb/btn-eepromdb.h>
#include <sys/event.h>
#include <sys/heap.h>
#include <sys/timer.h>
#include <sys/mutex.h>
#include <string.h>
#include <fcntl.h>
#include <bt/bt_hci_defs.h>
#include <hardware/btn-hardware.h>
#include <hardware/ram_banks.h>
#include <dev/usart.h>
#include <dev/usartavr.h>
#include <stdio.h>
#include <io.h>
#include <sys/tracer.h>
#include <debug/logging.h>

struct _log_cmds_stack_s {
	FILE *terminal_uart;
	log_filter_t verbosity;
};

struct _log_cmds_stack_s log_cmds_stack;

extern struct _log_stack log_stack;

// log level Names
static char logNameERROR[] =	"ERR: ";
static char logNameWARNING[] =	"WRN: ";
static char logNameINFO[] =		"INF: ";
static char logNameDEBUG[] =	"DBG: ";

static char* logLevelNames[LOG_LEVELS + 1] = {0,
										  logNameERROR,
                                          logNameWARNING,
                                          logNameINFO,
                                          logNameDEBUG};
                                          
void _log_cmds_set_verbosity( u_char class, u_char mask)
{
    if( class > 0 && class < LOG_MAX_CLASSES) {
        log_cmds_stack.verbosity[class] = mask;
    }
}

void _log_cmds_print_msg(log_msg_t* msg, void* cb_arg)
{
	va_list ap;
	log_msg_tail* tail = (log_msg_tail*)(msg->data + msg->len);
	char* fmt = tail->fmt;
	
	// get lock on terminal
    btn_terminal_get_writing_lock();
    // print timestamp
    fprintf( log_cmds_stack.terminal_uart,
    			"%4u:%03u ",
    			(u_short)((msg->t_stamp)/1000),
    			(u_short)((msg->t_stamp)%1000));
    // print log class
    fprintf(log_cmds_stack.terminal_uart, "C%.2d, ", msg->class);
    // print log level
    fprintf( log_cmds_stack.terminal_uart, "%s", logLevelNames[msg->level]);

    // print logged data - differ between strings and log messages
    if (fmt == NULL) {
    	fputs(msg->data, log_cmds_stack.terminal_uart);
    }
    else {
    	ap = (va_list)msg->data;
#ifdef LOG_USE_PROG_MEM
		vprintf_P(log_cmds_stack.terminal_uart, fmt, ap);
#else
    	vfprintf(log_cmds_stack.terminal_uart, fmt, ap);
#endif
    	va_end(ap);
    }
    // release lock on terminal
    btn_terminal_free_writing_lock();
}

void _log_cmds_print_usage(void)
{
    tprintf("log usage: log show [<nr_msgs> [<class> [<mask]]]\n");
    tprintf("           log clear\n");
    tprintf("           log logmask <class> [<mask>]\n");
    tprintf("           log verbosity <class> [<mask>]\n");
}

void _log_cmds_cmd(char * arg)
{
    u_int class = LOG_ALL_CLASSES, mask = LOG_MASK_ALL, nr_msgs = 0;
    int num;
    if (!strncmp(arg, "show", 4)) {
        sscanf(arg+4, "%u%u%x", &nr_msgs, &class, &mask);
        log_get_messages( nr_msgs, class, mask, _log_cmds_print_msg, NULL);
        return;
    } else if (!strncmp(arg, "clear", 5)) {
		log_empty();
        return;
    } else if (!strncmp(arg, "logmask", 7)) {
        num = sscanf(arg+7, "%d%x", &class, &mask);
        if( num == 2) {
            log_set_logmask( class, mask);
        }
        if( num >= 1) {
        	mask = log_get_logmask(class);
        	tprintf("log class %i, ", class);
        	if (mask < 0) {
        		tprintf("invalid!\n");
        		return;
        	}
        	else {
        		tprintf(" logmask set to: %.2X\n", mask);
        		return;
        	}
        }
    } else if (!strncmp(arg, "verbosity", 9)) {
        num = sscanf(arg+9, "%d%x", &class, &mask);
        if( num == 2) {
            _log_cmds_set_verbosity( class, mask);
        }
        if( num >= 1) {
            if( class > 0 && class < LOG_MAX_CLASSES) {
                tprintf("log class %i, verbosity mask set to: %.2X\n",
                                    class, log_cmds_stack.verbosity[class]);
                return;
            }
        }
    }
    _log_cmds_print_usage();
}

void log_cmds_init( FILE *uart)
{
	u_short i;
	short retval;
	
	// store pointer to the terminal uart
    log_cmds_stack.terminal_uart = uart;
	
	// init verbosity
	for (i=0; i<LOG_MAX_CLASSES; i++) {
		log_cmds_stack.verbosity[i] = LOG_VERBOSITY_OFF;
	}
	
	// register _log_cmds_print_msg() for verbosity
	retval = log_register_event_listener(_log_cmds_print_msg,
											NULL,
											log_cmds_stack.verbosity);
											
	// register cmds at terminal
	btn_terminal_register_cmd("log", _log_cmds_cmd);
}
