/*
 * Copyright (C) 2000-2005 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */


/**
 * \file support/rprog.c
 *
 * \date 14.04.2005
 *
 * \author Mustafa Yuecel <yuecelm@ee.ethz.ch>
 * \author Matthias Ringwald <mringwal@inf.ethz.ch>
 *
 * \brief btnode library to add remote programming cability (over L2CAP)
 *
 * \note using buffer handling from btnode/bt/bt_rfcomm.c
 *       Ole Reinhardt <ole.reinhardt@embedded-it.de>
 */
 
#include <stdio.h>
#include <string.h>
#include <sys/mutex.h>
#include <sys/timer.h>
#include <led/btn-led.h>
#include <cc/ccc.h>
#include <wsnlab/wsnlab.h> 

// ----------------------------------------------------------------------------
// #define constants
// ----------------------------------------------------------------------------

#define DICT_LENGTH 10

// ----------------------------------------------------------------------------
// local variables
// ----------------------------------------------------------------------------

static MUTEX dict_mutex;
static pkt_element_t packet_dict[DICT_LENGTH];

// ----------------------------------------------------------------------------
// helper functions
// ----------------------------------------------------------------------------

/**
 * Prints a string representation of the data to stdout (like a hex editor)
 */
void print_data( char* data, u_short length )
{
	u_short idx, jdx;
	
	for ( idx = 0; idx < length / 16; idx++ )
	{
		printf( "%04u   %02X %02X %02X %02X %02X %02X %02X %02X  "
				"%02X %02X %02X %02X %02X %02X %02X %02X", idx*16,
				data[0+idx*16], data[1+idx*16], data[2+idx*16], data[3+idx*16], 
				data[4+idx*16], data[5+idx*16], data[6+idx*16], data[7+idx*16],
				data[8+idx*16], data[9+idx*16], data[10+idx*16], data[11+idx*16], 
				data[12+idx*16], data[13+idx*16], data[14+idx*16], data[15+idx*16] );
				
		printf( "   %c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c\n", 
				data[0+idx*16], data[1+idx*16], data[2+idx*16], data[3+idx*16], 
				data[4+idx*16], data[5+idx*16], data[6+idx*16], data[7+idx*16],
				data[8+idx*16], data[9+idx*16], data[10+idx*16], data[11+idx*16], 
				data[12+idx*16], data[13+idx*16], data[14+idx*16], data[15+idx*16] );	
	}
	
	if ( length % 16 != 0 )
	{
		printf( "%04u   ", idx*16 );
		for ( jdx = idx*16; jdx < idx*16+8; jdx++ )
		{
			if ( jdx < length )
			{
				printf( "%02X ", data[jdx] );
			}
			else
			{
				printf( "   " );
			}				
		}	
		printf( " " );
		for ( ; jdx < (idx+1)*16; jdx++ )
		{
			if ( jdx < length )
			{
				printf( "%02X ", data[jdx] );
			}
			else
			{
				printf( "   " );
			}				
		}	
		printf( "  " );
		for ( jdx = idx*16; jdx < length; jdx++ )
		{
			printf( "%c", data[jdx] );
		}
		printf( "\n" );
	}
	printf( "Length of data block: %u bytes\n", length );
}

void print_packet_short( ccc_packet_t* pkt )
{
	if ( pkt != NULL )
	{
		printf( "%04X | %04X |   %02X |   %4u | %s\n",
		pkt->dst, pkt->src, pkt->type, pkt->length, pkt->data );
	}
	else
	{
		printf( " *** NO PACKET ***\n" );
	}
}

inline u_short pow2(u_short e)
{
    return 1 << e;
} 

// FIXME: This could be done with a single hw write operation
void led_set_binary(u_short i)
{
	u_short e=3;
	btn_led_clear(0);
	btn_led_clear(1);
	btn_led_clear(2);
	btn_led_clear(3);
	while (i)
	{
		while(i < pow2(e)) e--;
		btn_led_set(e);
		i -= pow2(e);
	}
}

// TODO: FIXME!
int ccc_send_packet( u_short dest, u_char type, u_char* data, u_short length )
{
	int res = 0;
	ccc_packet_t *pkt = new_ccc_packet( length );
	
	pkt->length = length;
	memcpy( pkt->data, data, length );
	
	res = ccc_send( dest, type, pkt );
	
	free( pkt );
	
	return res;
}

// ----------------------------------------------------------------------------
// Dictionary
// ----------------------------------------------------------------------------

/* Note: A dictionary has 3 operations: 
 * 1. is_member(k)
 * 2. insert(k)
 * 3. delete(k)
 */

void dict_init(void)
{
	int idx;
	
	NutMutexLock( &dict_mutex );		
	for ( idx = 0; idx < DICT_LENGTH; idx++ )
	{
		packet_dict[idx].send_counter = 0;
		packet_dict[idx].recv_counter = 0;		
		packet_dict[idx].received_at  = 0;
		packet_dict[idx].pkt = NULL;
	}
	NutMutexUnlock( &dict_mutex );			
}

void dict_print(void)
{
	int idx;
	
	NutMutexLock( &dict_mutex );		
	for ( idx = 0; idx < DICT_LENGTH; idx++ )
	{		
		printf( "#%02u | send: %u | recvd: %u | at: %lu | ago: %lu \n", 
		        idx, 
		        packet_dict[idx].send_counter, 
		        packet_dict[idx].recv_counter,
		        packet_dict[idx].received_at,
		        NutGetMillis() - packet_dict[idx].received_at );		        
		print_packet_short( packet_dict[idx].pkt );
	}
	NutMutexUnlock( &dict_mutex );			
}

int dict_packet_is_known( ccc_packet_t *pkt )
{
	NutMutexLock( &dict_mutex );
	int idx;
	
	for ( idx = 0; idx < DICT_LENGTH; idx++ )
	{
		if ( packet_dict[idx].pkt != NULL && 
		     memcmp( pkt->data, packet_dict[idx].pkt->data, pkt->length ) == 0 )
		{
			// update the receive counter
			packet_dict[idx].recv_counter++;
	
			printf( "Received duplicate packet at %u: ", idx );
			print_packet_short( pkt );
        
			NutMutexUnlock( &dict_mutex );
			//return idx;
			return 1;
		}
	}
	
	NutMutexUnlock( &dict_mutex );
	return 0;
}

int dict_add_packet( ccc_packet_t *pkt )
{
	NutMutexLock( &dict_mutex );
	
	int idx = 0;
	
	while ( idx < DICT_LENGTH && packet_dict[idx].pkt != NULL )
	{
		idx++;
	}
	
	if ( idx < DICT_LENGTH && packet_dict[idx].pkt == NULL )
	{
		// we found a place in the send queue, so we add the current packet
		packet_dict[idx].received_at  = NutGetMillis();
		packet_dict[idx].recv_counter = 0;
		packet_dict[idx].send_counter = 0;
		
		// copy the packet in a local buffer
		packet_dict[idx].pkt = new_ccc_packet( pkt->length );
		memcpy( packet_dict[idx].pkt, pkt, sizeof(ccc_packet_t) + pkt->length );		
	}
	
	NutMutexUnlock( &dict_mutex );	
	
	return idx;
}

void dict_remove_packet( int idx )
{
	packet_dict[idx].send_counter = 0;
	packet_dict[idx].recv_counter = 0;	
	packet_dict[idx].received_at = 0;
	free( packet_dict[idx].pkt );
	packet_dict[idx].pkt = NULL;
}

pkt_element_t* dict_get_first_packet( int* handle )
{
	int idx = 0;
	NutMutexLock( &dict_mutex );
	
	while ( idx < DICT_LENGTH && packet_dict[idx].pkt == NULL )
	{
		idx++;
	}
	
	if ( idx < DICT_LENGTH )
	{
		*handle = idx;
		NutMutexUnlock( &dict_mutex );			
		return &packet_dict[idx];
	}
	
	NutMutexUnlock( &dict_mutex );	
	
	return NULL;
}

pkt_element_t* dict_get_next_packet( int* handle )
{
	int idx = *handle + 1;	
	NutMutexLock( &dict_mutex );	
	
	while ( idx < DICT_LENGTH  && packet_dict[idx].pkt == NULL )
	{
		idx++;
	}
	
	if ( idx < DICT_LENGTH )
	{
		*handle = idx;
		NutMutexUnlock( &dict_mutex );			
		return &packet_dict[idx];
	}	
	
	NutMutexUnlock( &dict_mutex );		
	return NULL;
}
